/*
 * i200m.c
 *
 * Driver code for Intellisys 200 motor controller 
 *
 * Copyright(C) 2000 Nomadic Technologies, Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU Public License, incorporated herein by reference.
 *
 * Authors:
 * Rich LeGrand (rmlegran@robots.com) -- original author
 * Jake Sprouse (jake@robots.com) -- current maintainer
 *
 * $Log: i200m.c,v $
 * Revision 1.5  2000/03/13 20:06:00  jake
 * we now request a dynamically assigned major number when registering
 * as a character device
 *
 * Revision 1.4  2000/03/07 17:50:01  jake
 * added GPL reference to header
 *
 * Revision 1.3  2000/03/07 15:05:48  jake
 * changed the request queue mechanism to handle the case of a task going
 * away (as on a signal) before a request has been handled.  a device now
 * keeps a pointer to the request being serviced and a doubly-linked list
 * of pending requests.  if a task goes away while its request is pending,
 * the request is simply removed from the list.  if it goes away when
 * the request is being serviced, the wait queue of the request is set to
 * NULL so that the isr will not signal it when the request is complete.
 *
 * Revision 1.2  2000/02/22 21:44:18  jake
 * - made the read/write calls handle timeouts and signals better.  they now
 *   both call i200m_task_send_req, which does the following: (1) send the req
 *   to dpr if there are no request pending, or else adds it to the queue and
 *   sleeps until it is pending; (2) sleeps until it has been completely
 *   serviced; (3) if either sleep times out or is interrupted, the request
 *   is removed from the queue.  implementation of (3) required making the
 *   queue doubly-linked.
 * - fixed a bug where the isr was setting the servo event wait queues to
 *   NULL after waking them up; this was causing select() to crash the
 *   kernel.
 * - fixed a bug where the return value of copy_to/from_user was being
 *   misinterpreted; now, if they return non-zero, -EFAULT is returned.
 *
 * Revision 1.1  2000/01/16 01:01:52  jake
 * extensively cleaned up and rewritten from the original i200mdriver.c code
 * contained in the Nomad200 "robot" source tree.
 *
 */

#include <linux/module.h>
#include <linux/version.h>

char kernel_version[] = UTS_RELEASE;

#include <linux/wait.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/malloc.h>
#include <linux/poll.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include "i200m.h"



/* ideas:
 *
 * - use the mailboxes for function codes instead of the dpr
 */


#define I200M_MAJOR             52
#define I200M_BASE_ADDR         0xd2000

#define I200M_REQ_TIMEOUT  (2*HZ)  /* max time a request should take */


/* command codes for upload mode */
#define I200M_BOOTCMD_NOP    0
#define I200M_BOOTCMD_XFER   1
#define I200M_BOOTCMD_RUN    2
#define I200M_BOOTCMD_DONE   0xff

/* command codes for the function port */
#define I200M_CMD_READ_REG      0
#define I200M_CMD_WRITE_REG     1
#define I200M_CMD_READ_CON      2
#define I200M_CMD_WRITE_CON     3
#define I200M_CMD_RESET         255

/* interrupt event bits (interrupt_flags register) */
#define I200M_EVENT_READ           0x0001
#define I200M_EVENT_WRITE          0x0002
#define I200M_EVENT_SERVO_DONE     0x0004
#define I200M_EVENT_SERVO_ERROR    0x0008
#define I200M_EVENT_SERVO_BKPNT    0x0010
#define I200M_EVENT_POSN_WRAP      0x0020
#define I200M_EVENT_NEG_LIMIT      0x0040
#define I200M_EVENT_POS_LIMIT      0x0080
#define I200M_EVENT_READ_CONSOLE   0x0100
#define I200M_EVENT_WRITE_CONSOLE  0x0200

/* bitmask for the host status port */
#define I200M_HOST_READING      0x0001
#define I200M_HOST_BUSY         0x0002

/* max number of devices.  increment as needed */
#define I200M_NO                1

/* number of axes we're controlling */
#define I200M_NUM_AXES          8


/* debugging */
#ifdef I200M_DEBUG
#define PDEBUG(fmt, args...) printk("i200m: " fmt, ## args)
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#else
#define PDEBUG(fmt, args...)
#endif


/* this DPR layout is used by the boot code (before a device's state
 * is READY).  total memory size is 8K less the semaphore, version,
 * command, error, arg, and reserved fields, and two bytes for the
 * mailboxes */
struct i200m_dpr_boot
{
  unsigned char semaphore[8];
  unsigned char version[8];
  unsigned char memory[8164];
  unsigned short command;
  unsigned short error;
  unsigned long arg;
};



/* this is the convention used to divide up the onboard DPR */
struct i200m_dpr_layout
{
  unsigned char semaphore[8];
  unsigned short function;
  unsigned short reg;
  long data;
  short result;
  unsigned char console_read;
  unsigned char console_write;
  unsigned short interrupt_flags;
  unsigned short interrupt_mask;
  unsigned short status;
  unsigned short host_status;
  unsigned long servo_done;
  unsigned long servo_error;
  unsigned long servo_bkpnt;
};


/* each device has a FIFO queue of reqests which are executed in sequence */
struct i200m_request
{
  short function;
  char *buf;
  int len;
  int index;             /* a write req can have multiple reqs per buffer */
  struct wait_queue *wq; /* this will be signaled with the req is acked */
  int result;
  struct i200m_request *next, *last;
};



/* per-device data */
struct i200m_device
{
  unsigned short magic;
  volatile struct i200m_dpr_layout *dpr;
  volatile unsigned char *my_mbox, *her_mbox; /* mailboxes */
  int irq;
  int open_count;
  enum { UNKNOWN = 0, START_SREC, GET_SREC, BAD_SREC, READY } state;
  unsigned int srec_count, srec_total, bad_srec_count;
  unsigned int upload_index, upload_index0;
  struct i200m_request *req_queue_head, *req_queue_tail, *req_pending;
  struct wait_queue *servo_done_wqs[8], *servo_error_wqs[8],
    *servo_bkpnt_wqs[8];
};


/* per-file data */
struct i200m_file
{
  struct i200m_device *dev;
  unsigned long servo_done_mask;
  unsigned long servo_error_mask;
  unsigned long servo_bkpnt_mask;
};



/* the array of devices.  i200m_no_devs reflects the last used */
static struct i200m_device i200m_devices[I200M_NO];
static int i200m_no_devs = 0;

/* our major number, assigned dynamically by register_chrdev() */
static int i200m_major;



static int i200m_boot_command(struct i200m_device *dev, unsigned short cmd,
			      unsigned long arg)
{
  struct i200m_dpr_boot *boot = (struct i200m_dpr_boot *)dev->dpr;
  unsigned long t0;
  unsigned short result;

  /* store the command and interrupt */
  writew(htons(cmd), &boot->command);
  writel(htonl(arg), &boot->arg);
  writew(0, &boot->error);
  writeb(0x01, dev->her_mbox);

  t0 = jiffies;
  do
  {
    schedule();
    if (jiffies - t0 > 4*HZ) /* timeout after four seconds */
    {
      printk("i200m: boot command timeout\n");
      return -EIO;
    }
  } while ((result = ntohs(readw(&boot->command))) == cmd);

  /* wait for ten msec so we don't give the next command too fast */
  t0 = jiffies;
  do
  {
    schedule();
  } while (jiffies - t0 < HZ/100);

  if (result != I200M_BOOTCMD_DONE)
  {
    printk("i200m: unexpected boot command register 0x%x\n", (int)result);
    return -EIO;
  }

  if ((result = ntohs(readw(&boot->error))) == 0)
  {
    return 0;
  }

  printk("i200m: boot command error 0x%x\n", (int)result);
  return -EIO;
}



/* called when a sub-request is acknowledged.  returns 0 if there are more
 * subrequests to process. */
static int i200m_reqdone(struct i200m_device *dev, struct i200m_request *req)
{
  switch (req->function)
  {
  case I200M_CMD_READ_REG:

    *(unsigned long *)req->buf = ntohl(readl(&dev->dpr->data));
    req->result = ntohs(readw(&dev->dpr->result));
    return 1;

  case I200M_CMD_WRITE_REG:

    if (req->index >= req->len/sizeof(struct i200m_write_req))
    {
      req->result = ntohs(readw(&dev->dpr->result));
      return 1;
    }
    else
    {
      return 0;
    }
  }

  return 1;
}



/* send the next sub-request of a request. */
static void i200m_nextreq(struct i200m_device *dev)
{
  struct i200m_write_req *wreq;
  struct i200m_request *req = dev->req_pending;

  writew(htons(req->function), &dev->dpr->function);

  switch (req->function)
  {
  case I200M_CMD_READ_REG:

    writew(htons(*(unsigned short *)req->buf), &dev->dpr->reg);
    break;

  case I200M_CMD_WRITE_REG:

    wreq = &((struct i200m_write_req *)req->buf)[req->index];
    writew(htons(wreq->reg), &dev->dpr->reg);
    writel(htonl(wreq->value), &dev->dpr->data);
    break;
  }

  /* interrupt mc */
  writeb(0xff, dev->her_mbox);
  ++req->index;
  return;
}



static void i200m_unselect(struct wait_queue *wqs[I200M_NUM_AXES],
			   unsigned long mask)
{
  int i;

  for (i = 0; i < I200M_NUM_AXES; ++i)
  {
    if ((mask & 1) != 0 && wqs[i] != NULL)
    {
      wake_up_interruptible(&wqs[i]);
    }
    mask >>= 1;
  }
  return;
}



static void i200m_service_irq(int irq, void *dev_id, struct pt_regs *regs)
{
  struct i200m_device *dev = (struct i200m_device *)dev_id;
  short event_bits;

  if (dev == NULL)
  {
    printk("i200m_service_irq(): irq %d for unknown device\n", irq);
    return;
  }

  if (dev->magic != 0xbeef)
  {
    printk("i200m_service_irq(): no beef\n");
    return;
  }

  /* clear interrupt */
  readb(dev->my_mbox);

  event_bits = ntohs(readw(&dev->dpr->interrupt_flags));

  if ((event_bits & I200M_EVENT_READ) != 0 ||
      (event_bits & I200M_EVENT_WRITE) != 0)
  {
    if (dev->req_pending == NULL || i200m_reqdone(dev, dev->req_pending) != 0)
    {
      /* req_pending == NULL indicates this request was abandoned */
      if (dev->req_pending != NULL)
      {
	/* tell the requesting task that it's done */
	wake_up_interruptible(&dev->req_pending->wq);
      }

      /* get the next request from the queue and make it pending */
      if ((dev->req_pending = dev->req_queue_head) != NULL)
      {
	if ((dev->req_queue_head = dev->req_queue_head->next) != NULL)
	{
	  dev->req_queue_head->last = NULL;
	}
	else
	{
	  dev->req_queue_tail = NULL;
	}
      }
    }

    /* send the next request */
    if (dev->req_pending != NULL)
    {
      i200m_nextreq(dev);
    }

    event_bits &= ~(I200M_EVENT_READ | I200M_EVENT_WRITE);
  }

  if (event_bits & I200M_EVENT_SERVO_DONE)
  {	  
    i200m_unselect(dev->servo_done_wqs, ntohl(readl(&dev->dpr->servo_done)));
    event_bits &= ~I200M_EVENT_SERVO_DONE;
  }
  if (event_bits & I200M_EVENT_SERVO_ERROR)
  {	  
    i200m_unselect(dev->servo_error_wqs,
		   ntohl(readl(&dev->dpr->servo_error)));
    event_bits &= ~I200M_EVENT_SERVO_ERROR;
  }
  if (event_bits & I200M_EVENT_SERVO_BKPNT)
  {	  
    i200m_unselect(dev->servo_bkpnt_wqs,
		   ntohl(readl(&dev->dpr->servo_bkpnt)));
    event_bits &= ~I200M_EVENT_SERVO_BKPNT;
  }

  writew(htons(event_bits), &dev->dpr->interrupt_flags);
  return;
}



static int i200m_reset(struct i200m_device *dev)
{
  unsigned long start_jiffies;

  /* clear my mailbox */
  readb(dev->my_mbox);

  /* write reset function and interrupt */
  writew(htons(I200M_CMD_RESET), &dev->dpr->function);
  writeb(0xff, dev->her_mbox);

  /* wait for reset */
  start_jiffies = jiffies;
  while (jiffies - start_jiffies < HZ/4)
  {
    schedule();
  }

  /* turn interrupt flags off and clear host flag */
  writew(0x0000, &dev->dpr->interrupt_mask);
  writew(readw(&dev->dpr->host_status) & ~I200M_HOST_READING,
	 &dev->dpr->host_status);

  /* currently we assume that the device is only reset when the device
   * is not in the READY state, meaning that no processes are waiting
   * for requests to finish */
  dev->state = START_SREC;
  dev->upload_index = 0;
  dev->upload_index0 = 0;
  dev->srec_count = dev->srec_total = dev->bad_srec_count = 0;
  dev->req_queue_head = dev->req_queue_tail = dev->req_pending = NULL;

  return 0;
}



/* * this function is called by the read and write routines to process
 * request.  it queues the request and sleeps until the request's wait
 * queue has been signaled by the ISR, indicating completion */
static int i200m_task_send_req(struct i200m_device *dev,
			       struct i200m_request *req,
			       unsigned long timeout)
{
  int result = 0;

  /* check to see if mc is busy (request queue is empty) */
  cli();
  if (dev->req_pending == NULL)
  {
    /* no requests currently pending; we can send this request now */
    dev->req_pending = req;
    i200m_nextreq(dev);
  }
  else
  {
    /* add to the end of the queue and wait for it to be sent (isr signals
     * when it has given the request to the card) */
    req->next = NULL;
    if (dev->req_queue_tail != NULL)
    {
      req->last = dev->req_queue_tail;
      dev->req_queue_tail->next = req;
      dev->req_queue_tail = req;
    }
    else
    {
      dev->req_queue_head = dev->req_queue_tail = req;
      req->last = NULL;
    }
  }

  /* this request is now pending -- wait for the request to finish */
  req->wq = NULL;
  interruptible_sleep_on(&req->wq);
  if (signal_pending(current))
  {
    /* ### SMP alert!  we probably need spinlocks here to prevent a
     * ### second task from adding a new request to the req_queue as we
     * ### are removing this one... */
    if (dev->req_pending == req)
    {
      /* ack from card will be ignored by isr */
      dev->req_pending = NULL;
    }
    else
    {
      /* remove from req_queue */
      if (req->last != NULL)
      {
	req->last->next = req->next;
      }
      else
      {
	dev->req_queue_head = req->next;
      }
      if (req->next != NULL)
      {
	req->next->last = req->last;
      }
      else
      {
	dev->req_queue_tail = req->last;
      }
    }
    if (signal_pending(current))
    {
      result = -EINTR;
    }
    else
    {
      printk("i200m: request timeout\n");
      result = -EIO;
    }
  }

  sti();
  return result;
}



static ssize_t i200m_read(struct file *fp, char *buf,
			  size_t len, loff_t *off)
{
  struct i200m_file *priv = (struct i200m_file *)fp->private_data;
  struct i200m_request req;
  int result;
  char reqbuf[256];

  /* ### disallow interrupts... why??  doesn't cli()/sti() work?? */
  writew(0xffff, &priv->dev->dpr->interrupt_mask);

  if (len > sizeof(reqbuf))
  {
    req.buf = kmalloc(len, GFP_KERNEL);
  }
  else
  {
    req.buf = reqbuf;
  }
  if (copy_from_user(req.buf, buf, len) != 0)
  {
    return -EFAULT;
  }
  req.len = len;
  req.index = 0;
  req.function = I200M_CMD_READ_REG;

  /* send the request (or queue it) */
  if ((result = i200m_task_send_req(priv->dev, &req, I200M_REQ_TIMEOUT)) == 0)
  {
    if (copy_to_user(buf, req.buf, len) != 0)
    {
      result = -EFAULT;
    }
  }
  if (result == 0)
  {
    result = req.result;
  }
  if (req.buf != reqbuf)
  {
    kfree(req.buf);
  }

  return result;
}



static int i200m_writesrec(struct i200m_device *dev,
			   const char *buf, int len)
{
  struct i200m_dpr_boot *boot = (struct i200m_dpr_boot *)dev->dpr;
  int index = 0, result, i;
  char ch;
  const char *bp;

  while (index < len)
  {
    switch (dev->state)
    {
    case BAD_SREC:

      bp = &buf[index++];
      if ((result = get_user(ch, bp)) != 0)
      {
	return result;
      }
      if (ch == '\n')
      {
	dev->upload_index = dev->upload_index0;
	dev->state = START_SREC;
      }
      continue;

    case START_SREC:

      bp = &buf[index++];
      if ((result = get_user(ch, bp)) != 0)
      {
	return result;
      }
      if (ch != 'S')
      {
	++dev->bad_srec_count;
	dev->state = BAD_SREC;
	continue;
      }
      else
      {
	writeb('S',  boot->memory + dev->upload_index);
	dev->state = GET_SREC;
      }
      break;

    case GET_SREC:

      bp = &buf[index++];
      if ((result = get_user(ch, bp)) != 0)
      {
	return result;
      }
      if (ch == '\r')
      {
	continue;
      }
      if (ch == '\n')
      {
	ch = '\0';
      }
      if (ch != '\0' && (ch < '0' || (ch > '9' && ch < 'A') || ch > 'F'))
      {
	++dev->bad_srec_count;
	dev->state = BAD_SREC;
	continue;
      }
      writeb(ch, boot->memory + dev->upload_index);
      if (ch == '\0')
      {
	++dev->srec_count;
	dev->state = START_SREC;
	dev->upload_index0 = dev->upload_index + 1;
      }
      break;

    default:
      PDEBUG("bad state in writesrec!!\n");
      return -EIO;
    }

    if (++dev->upload_index >= sizeof(boot->memory))
    {
      if (dev->srec_count == 0)
      {
	/* ERROR -- huge SREC */
	return -EFAULT;
      }

      result = i200m_boot_command(dev, I200M_BOOTCMD_XFER, dev->srec_count);
      if (result < 0)
      {
	/* command timed out */
	i200m_reset(dev);
	return result;
      }

      dev->srec_total += dev->srec_count;
      dev->srec_count = 0;
      if (dev->upload_index > dev->upload_index0)
      {
	for (i = 0; i < dev->upload_index - dev->upload_index0; ++i)
	{
	  writeb(readb(&boot->memory[dev->upload_index0 + i]),
		 &boot->memory[i]);
	}
      }
      dev->upload_index -= dev->upload_index0;
      dev->upload_index0 = 0;
    }
  }

  return index;
}



static ssize_t i200m_write(struct file *fp, const char *buf,
			   size_t len, loff_t *off)
{
  struct i200m_file *priv = (struct i200m_file *)fp->private_data;
  struct i200m_request req;
  char reqbuf[256];
  int result;

  if (priv->dev->state != READY)
  {
    /* still in upload mode */
    return i200m_writesrec(priv->dev, buf, len);
  }

  /* ### disallow interrupts... why??  doesn't cli()/sti() work?? */
  writew(0xffff, &priv->dev->dpr->interrupt_mask);

  if (len > sizeof(reqbuf))
  {
    req.buf = kmalloc(len, GFP_KERNEL);
  }
  else
  {
    req.buf = reqbuf;
  }
  if (copy_from_user(req.buf, buf, len) != 0)
  {
    return -EFAULT;
  }
  req.len = len;
  req.index = 0;
  req.function = I200M_CMD_WRITE_REG;

  /* send the request (or queue it) */
  if ((result = i200m_task_send_req(priv->dev, &req,
       I200M_REQ_TIMEOUT*len/sizeof(struct i200m_write_req))) == 0)
  {
    result = req.result;
  }
  if (req.buf != reqbuf)
  {
    kfree(req.buf);
  }

  return result;
}



static void i200m_addpoll(struct file *fp, unsigned long mask,
			  struct wait_queue *wqs[I200M_NUM_AXES],
			  poll_table *table)
{
  int i;

  for (i = 0; i < I200M_NUM_AXES; ++i)
  {
    if ((mask & 1) != 0)
    {
      poll_wait(fp, &wqs[i], table);
    }
    mask >>= 1;
  }

  return;
}



static unsigned int i200m_poll(struct file *fp, poll_table *table)
{
  struct i200m_file *priv = (struct i200m_file *)fp->private_data;
  int event;

  /* ### disallow interrupts... why?? */
  writew(0xffff, &priv->dev->dpr->interrupt_mask);

  i200m_addpoll(fp, priv->servo_done_mask, priv->dev->servo_done_wqs, table);
  i200m_addpoll(fp, priv->servo_error_mask, priv->dev->servo_error_wqs, table);
  i200m_addpoll(fp, priv->servo_bkpnt_mask, priv->dev->servo_bkpnt_wqs, table);

  cli();
  event = (readl(&priv->dev->dpr->servo_done) & priv->servo_done_mask) != 0 ||
    (readl(&priv->dev->dpr->servo_error) & priv->servo_error_mask) != 0 ||
    (readl(&priv->dev->dpr->servo_bkpnt) & priv->servo_bkpnt_mask) != 0;
  sti();

  return event ? POLLIN | POLLRDNORM : 0;
}




static int i200m_ioctl(struct inode *in, struct file *fp, unsigned int cmd, 
		       unsigned long value)
{
  struct i200m_file *priv = (struct i200m_file *)fp->private_data;

  switch (cmd)
  {
  case I200M_IOCDONEMASK:
    priv->servo_done_mask = value;
    break;

  case I200M_IOCERRMASK:
    priv->servo_error_mask = value;
    break;

  case I200M_IOCBRKMASK:
    priv->servo_bkpnt_mask = value;
    break;

  default:
    return -EINVAL;
  }

  return 0;
}




static int i200m_open(struct inode *in, struct file *fp)
{
  struct i200m_device *dev;
  struct i200m_file *priv;
  int result;

  if (MINOR(in->i_rdev) >= i200m_no_devs)
  {
    return -ENODEV;
  }
  
  dev = &i200m_devices[MINOR(in->i_rdev)];

  priv = (struct i200m_file *)kmalloc(sizeof(struct i200m_file), GFP_KERNEL);
  if (priv == NULL)
  {
    return -ENOMEM;
  }
  memset(priv, 0, sizeof(struct i200m_file));
  priv->dev = dev;
  fp->private_data = priv;

  /* we enable interrupts if the embedded code is ready */
  if (dev->state == READY)
  {
    /* if this is the first open, register the ISR */
    if (dev->open_count == 0)
    {
      result = request_irq(dev->irq, i200m_service_irq,
			   SA_INTERRUPT, "i200m", dev);
      if (result < 0)
      {
	printk("i200m: unable to register interrupt %d\n", dev->irq);
	return result;
      }
      PDEBUG("registered irq #%d\n", dev->irq);
    }
  }
  else
  {
    /* only one process can open us while we're bootstrapping */
    if (dev->open_count > 0)
    {
      return -EBUSY;
    }
  }
  
  ++dev->open_count;
  MOD_INC_USE_COUNT;
  return 0;
}



static int i200m_release(struct inode *in, struct file *fp)
{
  struct i200m_device *dev;
  
  if (MINOR(in->i_rdev) >= i200m_no_devs)
  {
    return -ENODEV;
  }

  dev = &i200m_devices[MINOR(in->i_rdev)];

  MOD_DEC_USE_COUNT;
  --dev->open_count;

  if (dev->state != READY)
  {
    if (dev->state == START_SREC)
    {
      if (dev->srec_count > 0)
      {
	/* transfer remaining s-records */
	if (i200m_boot_command(dev, I200M_BOOTCMD_XFER,
			       dev->srec_count) < 0)
	{
	  /* command timed out */
	  i200m_reset(dev);
	  return 0;
	}
      }

      dev->srec_total += dev->srec_count;
      if (dev->srec_total > 0)
      {
	if (dev->bad_srec_count > 0)
	{
	  printk("i200m: there were %d bad s-records in the uploaded file\n",
		dev->bad_srec_count);
	}
	
	/* start the program... */
	if (i200m_boot_command(dev, I200M_BOOTCMD_RUN, 0x100000) < 0)
	{
	  i200m_reset(dev);
	  return 0;
	}
	dev->state = READY;
	printk(KERN_INFO "i200m: uploaded %d s-records; device is ready\n",
	       dev->srec_total);
      }
    }
    else
    {
      printk("i200m: unexpected EOF in S-record upload.\n");
      i200m_reset(dev);
    }
  }
  else
  {
    /* release the IRQ if there's no more processes accessing us */
    if (dev->open_count == 0)
    {
      free_irq(dev->irq, dev);
    }
  }

  kfree(fp->private_data);

  return 0;
}



static struct file_operations i200m_fops =
{
  NULL, /* seek */
  i200m_read,
  i200m_write,
  NULL, /* readdir */
  i200m_poll,
  i200m_ioctl,
  NULL, /* mmap */
  i200m_open,
  NULL, /* flush */
  i200m_release,
  NULL, /* fsync */
  NULL, /* fasync */
  NULL, /* check_media_change */
  NULL, /* revalidate */
  NULL  /* lock */
};



/* --- /proc filesystem interface --- */



static int i200m_read_proc(char *buf, char **start, off_t offset,
			   int len, int unused)
{
  int i;
  struct i200m_device *dev;

  len = 0;
  for (i = 0; i < i200m_no_devs; ++i)
  {
    dev = &i200m_devices[i];
    len += sprintf(buf + len, "%d: open:%d state:", i, dev->open_count);
    switch (dev->state)
    {
    case UNKNOWN:
      len += sprintf(buf + len, "unknown");
      break;
    case START_SREC:
      len += sprintf(buf + len, "start_srec");
      break;
    case GET_SREC:
      len += sprintf(buf + len, "get_srec");
      break;
    case BAD_SREC:
      len += sprintf(buf + len, "bad_srec");
      break;
    case READY:
      len += sprintf(buf + len, "ready");
      break;
    }
    len += sprintf(buf + len, "\n");
  }

  return len;
}



static struct proc_dir_entry i200m_proc_entry =
{
  0,                 /* low_ino: the inode -- dynamic */
  5, "i200m",        /* length of name, and name */
  S_IFREG | S_IRUGO, /* mode */
  1, 0, 0,           /* nlinks, owner, group */
  0,                 /* size -- unused */
  NULL,              /* operations -- use default */
  &i200m_read_proc,  /* read function */
};




/* --- initialization routines --- */



static inline int release_sem(struct i200m_device *dev, unsigned char num)
{
  writeb(0xff, &dev->dpr->semaphore[num]);
  return (readb(&dev->dpr->semaphore[num]) == 0xff) ? 0 : -1;
}



static inline int test_get_sem(struct i200m_device *dev, unsigned char num)
{
  writeb(0xfe, &dev->dpr->semaphore[num]);
  return (readb(&dev->dpr->semaphore[num]) == 0x00) ? 0 : -1;
}



/* initialization function; called at boot time or, if compiled as a
 * module, at load time */
static int i200m_probe(struct i200m_device *dev)
{
  int i;
  unsigned char val;

  /* the only read-only probing we can really do is to check the
   * memory locations of the eight semaphores to see if any of them return
   * anything but 0x00 or 0xff.  if so, we know there's something else
   * there. */
  for (i = 0; i < 8; ++i)
  {
    if ((val = readb(&dev->dpr->semaphore[i])) != 0xff && val != 0x00)
    {
      printk("i200m: device not found at 0x%p (location of semaphore %d "
	     "returned 0x%d)\n", (char *)&dev->dpr, i, (int)val);
      return -ENODEV;
    }
  }

  /* to verify that we're looking at a memnet, we release, grab, and
   * re-release each semaphore.  we use val to hold the count of
   * semaphores which we successfully grab */
  val = 0;
  for (i = 0; i < 8; i++)
  {
    if (release_sem(dev, i) != 0)
    {
      printk("i200m: device not found at 0x%p\n", (char *)&dev->dpr);
      if (i > 0)
      {
	printk("i200m: initial release of semaphore %d failed\n", i);
      }
      return -ENODEV;
    }
    if (test_get_sem(dev, i) == 0)
    {
      ++val;
    }
    if (release_sem(dev, i) != 0)
    {
      printk("i200m: device not found at 0x%p\n", (char *)&dev->dpr);
      if (i > 0)
      {
	printk("i200m: release of semaphore %d failed\n", i);
      }
      return -ENODEV;
    }
  }

  /* check that we were able to grab at least one semaphore.  we assume
   * that the other side is not holding all of them, and if we got here
   * without grabbing any of them we must be looking at ROM set to 0xff */
  if (val == 0)
  {
    printk("i200m: device not found at 0x%p (no semaphores responded)\n",
	   (char *)&dev->dpr);
    return -ENODEV;
  }

  return 0;
}






int init_module(void)
{
  struct i200m_device *dev;
  int result, i;

  if ((result = register_chrdev(0, "i200m", &i200m_fops)) < 0)
  {
    printk("i200m: unable to register i200m device (%d)\n", result);
    return result;
  }
  i200m_major = result;
  printk(KERN_INFO "i200m: major number is %d\n", i200m_major);

  i200m_no_devs = 0;

  /* instanciate the singleton device */
  dev = &i200m_devices[0];
  dev->magic = 0xbeef;
  dev->dpr = (struct i200m_dpr_layout *)I200M_BASE_ADDR;
  dev->my_mbox = (unsigned char *)(I200M_BASE_ADDR + 0x1ffe);
  dev->her_mbox = (unsigned char *)(I200M_BASE_ADDR + 0x1fff);
  dev->irq = 7;
  for (i = 0; i < I200M_NUM_AXES; ++i)
  {
    dev->servo_done_wqs[i] = NULL;
    dev->servo_error_wqs[i] = NULL;
    dev->servo_bkpnt_wqs[i] = NULL;
  }
  dev->open_count = 0;

  if ((result = i200m_probe(dev)) != 0)
  {
    return result;
  }

  if ((result = i200m_reset(dev)) == 0)
  {
    ++i200m_no_devs;
  }

  proc_register(&proc_root, &i200m_proc_entry);
  return 0;
}



void cleanup_module(void)
{
  proc_unregister(&proc_root, i200m_proc_entry.low_ino);
  unregister_chrdev(i200m_major, "i200m");
  return;
}
