#ifndef _LINUX_I200M_H
#define _LINUX_I200M_H


/* ioctls to set up select.  when select() is done for reading, the process
 * will block until an event happens on one of the specified axes.  "motion
 * done", "motion error", and "breakpoint reached" events can be specified.
 * the argument to the ioctl should be a mask where each bit corresponds
 * to an axis, and an asserted bit indicates that that axis should be
 * watched */
#define I200M_IOCDONEMASK     0 
#define I200M_IOCERRMASK      1
#define I200M_IOCBRKMASK      2


/* the following registers are global (non-axis-specific) and do not need
 * an axis argument, although they should still be register mapped */
#define I200M_REG_CONSOLE_MODE               1
#define I200M_REG_VERSION                    2
#define I200M_REG_AXIS_ERROR                 3
#define I200M_REG_AXIS_DONE                  4
#define I200M_REG_AXIS_BKPNT                 5
#define I200M_REG_INTEGRALS                  6
#define I200M_REG_ESTOP                      7
#define I200M_REG_JOYSTICK_MODE              8
#define I200M_REG_JOYSTICK_STATE             9
#define I200M_REG_JOYSTICK_BUTTONS           10
#define I200M_REG_JOYSTICK_X_POT             11
#define I200M_REG_JOYSTICK_Y_POT             12
#define I200M_REG_ZEROING_AXES               13
#define I200M_REG_ZEROED_AXES                14
#define I200M_REG_BRAKE                      15
#define I200M_REG_BRAKE_MODE                 16

/* per axis registers */
#define I200M_REG_P_GAIN                     100
#define I200M_REG_I_GAIN                     101
#define I200M_REG_D_GAIN                     102
#define I200M_REG_VFF_GAIN                   103
#define I200M_REG_MOTOR_BIAS                 104
#define I200M_REG_MOTOR_LIMIT                105
#define I200M_REG_POSITION_ERROR_LIMIT       106
#define I200M_REG_INTEGRAL_LIMIT             107
#define I200M_REG_INTEGRAL_VALUE             108
#define I200M_REG_DEST_POSITION              109
#define I200M_REG_MAX_VELOCITY               110
#define I200M_REG_JERK                       111
#define I200M_REG_GEAR_RATIO                 112
#define I200M_REG_MASTER_AXIS                113
#define I200M_REG_PWM_VALUE                  114
#define I200M_REG_STATUS                     115
#define I200M_REG_ENCODER_VALUE              116
#define I200M_REG_PROFILE_MODE               117
#define I200M_REG_ZERO_MODE                  118
#define I200M_REG_BRAKE_MASK                 119
#define I200M_REG_ACCELERATION               120
#define I200M_REG_MAX_ACCELERATION           121
#define I200M_REG_ERROR_MODE                 122
#define I200M_REG_TARGET_VELOCITY            123
#define I200M_REG_TARGET_POSITION            124
#define I200M_REG_ZERO_STATE                 125
#define I200M_REG_LIMIT_SWITCH               126

/* per integral registers */
#define I200M_REG_ROTATE_ENC_PER_DEGREE      200
#define I200M_REG_ROTATE_AXIS                201
#define I200M_REG_TRANS_AXIS                 202
#define I200M_REG_X_INTEGRAL                 203
#define I200M_REG_Y_INTEGRAL                 204
#define I200M_REG_ROTATE_ANGLE               205

/* commands */
#define I200M_COM_GO                         500
#define I200M_COM_STOP_AXES                  501
#define I200M_COM_RECOVER                    502
#define I200M_COM_ZERO_INTEGRALS             504
#define I200M_COM_RESET                      505
#define I200M_COM_START_INTEGRALS            506
#define I200M_COM_STOP_INTEGRALS             507

/* zeroing modes */
#define ZERO_NEG_LIMIT                       0
#define ZERO_POS_LIMIT                       1
#define ZERO_HOME                            2

/* axis status bits */
#define I200M_AXIS_ERR_MOTION                0x01
#define I200M_AXIS_ERR_POS_LMT_SWITCH        0x02
#define I200M_AXIS_ERR_NEG_LMT_SWITCH        0x04
#define I200M_AXIS_ERR_JOYSTICKING           0x08
#define I200M_AXIS_ERR_ESTOP                 0x10

/* profile modes */
#define I200M_PROF_TRAP                      0
#define I200M_PROF_S_CURVE                   1
#define I200M_PROF_VELOCITY                  2
#define I200M_PROF_PWM                       3
#define I200M_PROF_E_GEAR                    4
#define I200M_PROF_ZERO                      5

/* joystick button masks -- to be used with I200M_REG_JOYSTICK_BUTTONS
   register */
#define I200M_JS_LEFT_MASK                   0x40
#define I200M_JS_TOP_MASK                    0x80

/* zeroing states */
#define I200M_ZS_START    0
#define I200M_ZS_1        1  /* initial go toward zero mark */
#define I200M_ZS_2        2  /* found zero mark -- go back to initial pos */
#define I200M_ZS_DONE     3


/* return results -- returned in result register */
#define I200M_RES_OK                         0
#define I200M_RES_ERROR                      -1
#define I200M_RES_INVALID_REG                -2

#define I200M_LIM_POS_MASK                 0x01
#define I200M_LIM_NEG_MASK                 0x02

/* the axis number to use in register maps for global registers */
#define I200M_GLOBAL                         0

/* read() and write() both require a register or command number.  also, if
 * the register or command operates on a specific axis (or dimenstion in
 * the case of integrals), it requires an axis number.  both of these
 * arguments are given in a 32-bit "register map", where the axis number is
 * specified in the lower five bits and the register number uses the
 * remaining 27.  the following macro generates a register map */
#define I200M_REG_MAP(r, ax) (((unsigned short)r << 5) + (unsigned short)ax)

/* write() to the motor controller should give a pointer to a buffer which
 * contains an array of the following struct, where the length of the buffer
 * is a multiple of sizeof(struct i200m_write_req) */
struct i200m_write_req
{
  unsigned short reg;  /* a register map (register or command plus axis) */
  long value;          /* value for register, or argument for command */
};



#endif /* _LINUX_I200M_H */
