/*
 * i200mdiag.c
 *
 * Low level interface utility for i200m devices
 *
 * Copyright(C) 2000 Nomadic Technologies, Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU Public License, incorporated herein by reference.
 *
 * Author: Jake Sprouse (jake@robots.com)
 *
 * $Log: i200minit.c,v $
 * Revision 1.3  2000/03/13 20:06:53  jake
 * added the -q flag for less verbose output
 *
 * Revision 1.2  2000/03/07 17:50:01  jake
 * added GPL reference to header
 *
 * Revision 1.1  2000/03/07 15:07:40  jake
 * initial commit; this is a program to upload embedded code to the
 * i200m module and then check the /proc/i200m file to make sure that
 * it was successfully started.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define DEFAULT_EMPROG   "/usr/local/share/i200m/i200m.prg"
#define DEFAULT_DEVBASE  "/dev/i200m"
#define DEFAULT_PROCFILE "/proc/i200m"


static char *argv0;


static void usage(int ret)
{
  fprintf((ret == 0) ? stdout : stderr, "usage: %s --help\n"
	  "%s [-f progfile]\n", argv0, argv0);
  exit(ret);
}



int main(int argc, char *argv[])
{
  int i = 0, line_no = 0, dev_no, open_count, quiet = 0;
  FILE *proc, *dev, *embed = NULL;
  const char *embed_file = DEFAULT_EMPROG, *dev_base = DEFAULT_DEVBASE,
    *proc_file = DEFAULT_PROCFILE;
  char dev_file[30], line[80], state[15], dbuf[1024];

  if ((argv0 = strrchr(argv[0], '/')) == NULL)
  {
    argv0 = argv[0];
  }
  else
  {
    ++argv0;
  }

  while (++i < argc)
  {
    if (strcmp(argv[i], "--help") == 0)
    {
      usage(0);
    }
    else if (strcmp(argv[i], "-f") == 0)
    {
      if (++i >= argc)
      {
	usage(-1);
      }
      embed_file = argv[i];
    }
    else if (strcmp(argv[i], "-p") == 0)
    {
      if (++i >= argc)
      {
	usage(-1);
      }
      proc_file = argv[i];
    }
    else if (strcmp(argv[i], "-d") == 0)
    {
      if (++i >= argc)
      {
	usage(-1);
      }
      dev_base = argv[i];
    }
    else if (strcmp(argv[i], "-q") == 0)
    {
      quiet = 1;
    }
    else
    {
      usage(-1);
    }
  }

  /* open the proc file */
  if ((proc = fopen(proc_file, "r")) == NULL)
  {
    fprintf(stderr, "%s: error opening %s for reading: %s\n",
	    argv0, proc_file, strerror(errno));
    exit(-1);
  }

  while (fgets(line, sizeof(line), proc) != NULL)
  {
    ++line_no;
    if ((i = sscanf(line, "%d: open:%d state:%s",
	       &dev_no, &open_count, state)) != 3)
    {
      fprintf(stderr, "%s: formatting mismatch on line %d of %s\n",
	      argv0, line_no, proc_file);
      exit(-1);
    }

    sprintf(dev_file, "%s%d", dev_base, dev_no);

    if (strcmp(state, "ready") == 0)
    {
      continue;
    }

    if (open_count != 0)
    {
      fprintf(stderr, "%s: device in use, can't upload to %s\n",
	      argv0, dev_file);
      continue;
    }

    if (strcmp(state, "start_srec") != 0)
    {
      fprintf(stderr, "%s: can't upload to %s because state is not "
	      "\"start_srec\"\n", argv0, dev_file);
      continue;
    }

    if ((dev = fopen(dev_file, "w")) == NULL)
    {
      fprintf(stderr, "%s: error opening %s for writing: %s\n",
	      argv0, dev_file, strerror(errno));
      continue;
    }

    if (embed == NULL)
    {
      if ((embed = fopen(embed_file, "r")) == NULL)
      {
	fprintf(stderr, "%s: error opening %s for reading: %s\n",
		argv0, embed_file, strerror(errno));
	exit(-1);
      }
    }

    while ((i = fread(dbuf, 1, sizeof(dbuf), embed)) > 0)
    {
      if (fwrite(dbuf, 1, i, dev) != i)
      {
	if (!ferror(dev))
	{
	  fprintf(stderr, "%s: unexpected short write to %s\n",
		  argv0, dev_file);
	}
	else
	{
	  fprintf(stderr, "%s: error while writing to %s: %s\n", argv[0],
		  dev_file, strerror(errno));
	}
	break;
      }
    }

    if (ferror(embed))
    {
      fprintf(stderr, "%s: error while reading %s: %s\n", argv[0], embed_file,
	      strerror(errno));
      exit(-1);
    }

    rewind(embed);

    if (quiet == 0)
    {
      printf("%s: uploaded %s to %s\n", argv0, embed_file, dev_file);
    }
    fclose(dev);
  }

  if (ferror(proc))
  {
    fprintf(stderr, "%s: error while reading %s: %s\n", argv[0], proc_file,
	    strerror(errno));
    exit(-1);
  }

  if (embed != NULL)
  {
    fclose(embed);
  }
  fclose(proc);

  return 0;
}
